/* -*- c -*- */
/@
   The KMAAB module shows the piece of memory in the kernel space 
   that is the MAAB-CARD PCI-Memory and can be mapped to the user memory.

   The driver search the PCI bus for MAAB Card and get the memory
   address of the card. In user space this memory area can be used for
   reading and writing

   The 'procinfo' method is used to make this buffer visible in the
   /proc/maab file.
   
      cat /proc/maab

   shows   
      Proc entry from maab:
      'Memory Content'
   
   The 'mmap' method maps the DSP memory space in user space
   see maablib for details.

@/

module "maab" : 
driver "maab" [ major=34 ] {


/*-------------------------------------------------*/

  file "globals" /{

/*
 * Copyright (c) 1998 IEM
 *
 * MAAB Globals and help files for configuration and Memory registers
 *
 * purpose: test for MAAB PCI-Card and get memory-space of DSP
 *
 */
    maab_dev maab; /* maab device nfo */

  }/


/*-------------------------------------------------*/
header "maab"/{

#define MAAB_VERBOSE 1

#if MAAB_VERBOSE
#define maab_says(x) {x;}
#else
#define maab_say(x) {}
#endif

#define MAAB_VendorId 0x1057 /* Motorola */
#define MAAB_DeviceId 0x1801 /* DSP 56301 */

#define MAAB_IO_SIZE 0x10000l

/* MAAB init routine */
#define MAAB_OK           0
#define MAAB_ERROR       -1
#define MAAB_NO_PCI      -2
#define MAAB_CANNOT_FIND -3
#define MAAB_IO_MAPPED   -4
  	
#if LINUX_VERSION_CODE < 0x020100                                     
#define ioremap vremap
#define iounmap vfree
#endif

typedef struct _maab_dev{

  /* for finding and reference */
  int pci_index;                /* which number */
  unsigned char pci_bus;        /* which bus */
  unsigned char pci_device_fn;  /* dont know */

  /* pxi interface */
  unsigned int pci_base_addr; /* phys. pci address */
  unsigned int pci_mem_type;  /* type of mem */
  char *pci_addr_memtype;     /* as ascii string */
  unsigned char pci_irq_line; /* irqline ? */
  char *pci_addr_prefetch ;   /* as ascii string */

  /* PCI Parameters for info */
  unsigned char pci_rev;            /* PCI_REVISION_ID */
  unsigned char pci_class_prog;     /* PCI_CLASS_PROG */
  unsigned char pci_class;           /* PCI_CLASS_DEVICE */

  unsigned char pci_cache_size;     /* PCI_CACHE_LINE_SIZE */
  unsigned char pci_latency;        /* PCI_LATENCY_TIMER */

  unsigned char pci_header_type;    /* PCI_HEADER_TYPE */
  unsigned char pci_bist;           /* PCI_BIST */

  unsigned int pci_status;          /*PCI_STATUS
					PCI_STATUS_66MHZ
					PCI_STATUS_UDF */

  /* remaped memory , bus addr. */
  unsigned char *io_base_addr;
  unsigned int io_size;
  
} maab_dev; /*struct name */


}/ /* header file */

/*-------------------------------------------------*/
core {  

init /{ 

#include <linux/bios32.h>  /* pcibios_ ... */
#include <linux/pci.h>     /* kernel PCI Defs */
  
  extern maab_dev maab;

  int retval = MAAB_CANNOT_FIND;

  printk("maab initialize: ");

  if(!pcibios_present()){
    printk("maab: No PCI Bios, init failed \n");
    return MAAB_NO_PCI;
  }

  /* search in each slot for maab */
  for(maab.pci_index = 0; maab.pci_index < 0x7f; maab.pci_index++){
    
    /* device_fn */
    if(pcibios_find_device(MAAB_VendorId,MAAB_DeviceId, maab.pci_index, 
			   &maab.pci_bus, &maab.pci_device_fn) !=0)
      break; /* try next slot */

    pcibios_read_config_byte(maab.pci_bus,maab.pci_device_fn, 
			     PCI_INTERRUPT_LINE,&maab.pci_irq_line);

    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn,
			     PCI_REVISION_ID ,
			     &maab.pci_rev);
    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn,
			     PCI_CLASS_PROG ,
			     &maab.pci_class_prog);
    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn, 
			     PCI_CLASS_DEVICE,
			     &maab.pci_class);
    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn,
			     PCI_CACHE_LINE_SIZE,
			     &maab.pci_cache_size);
    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn,
			     PCI_LATENCY_TIMER,
			     &maab.pci_latency);
    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn,
			     PCI_HEADER_TYPE,
			     &maab.pci_header_type);
    pcibios_read_config_byte(maab.pci_bus, maab.pci_device_fn,
			     PCI_BIST,
			     &maab.pci_bist);

    pcibios_read_config_dword(maab.pci_bus, maab.pci_device_fn,
			      PCI_STATUS,
			      &maab.pci_status);
    pcibios_read_config_dword(maab.pci_bus, maab.pci_device_fn,
			      PCI_BASE_ADDRESS_0,
			      &maab.pci_base_addr);

    if((maab.pci_base_addr & PCI_BASE_ADDRESS_SPACE_IO)!= 0){
      printk("maab found:but MAAB MEM is i/O ??? -> maab try next \n");
      break;
      /*      return MAAB_IO_MAPPED; */
    }

    /* process info dtat */
    switch((maab.pci_base_addr & PCI_BASE_ADDRESS_MEM_TYPE_MASK)){

    case PCI_BASE_ADDRESS_MEM_TYPE_32:
      maab.pci_addr_memtype = "PCI MEM_TYPE 32Bit";
      break;

    case PCI_BASE_ADDRESS_MEM_TYPE_1M:
      maab.pci_addr_memtype = "PCI MEM_TYPE 1M";
      break;

    case PCI_BASE_ADDRESS_MEM_TYPE_64:
      maab.pci_addr_memtype = "PCI MEM_TYPE 64Bit";
      break;
    }

    if((maab.pci_base_addr & PCI_BASE_ADDRESS_MEM_PREFETCH) == 0x0)
      maab.pci_addr_prefetch = "PCI MEM_PREFETCH DISABLED";
    else
      maab.pci_addr_prefetch = "PCI MEM_PREFETCH ENEABLED";

    /* mask address */
    maab.pci_base_addr &= PCI_BASE_ADDRESS_MEM_MASK; 

    maab.io_size = MAAB_IO_SIZE; /* see DSP 563011 manual */
    maab.io_base_addr = ioremap(maab.pci_base_addr,maab.io_size);

    /* sucess */
    printk("maab found MAAB at bus %d index %d with irq-line %d,",
	   maab.pci_bus, maab.pci_index, maab.pci_irq_line);
    printk("addr 0x%x\n",maab.pci_base_addr);

    /*return MAAB_OK;*/
    retval = MAAB_OK;

  }; /* for each maab.pci_index */
  
  
  if(retval != MAAB_OK){
    printk("maab: MAAB Card not found, init failed \n");
    return MAAB_CANNOT_FIND;
  }
			
}/

cleanup /{

  extern maab_dev maab;

  iounmap(maab.io_base_addr); /* free mapping */
  maab_says(printk("mass says Good bye! \n"));
			
}/
	   } /*core */
/*-------------------------------------------------*/

/*
 * This method will be called whenever the device is opened
 */

open /{
  extern maab_dev maab;

  maab_says(printk("%<driver.name> module has been opened\n"));
}/

/*
 * This method will be called whenever the device is closed
 */

close /{
  extern maab_dev maab;

  maab_says(printk("maab module has been closed ! \n"));
}/


/*
 * This method will be called whenever the opened deveice is mapped
 */

mmap /{

#include <linux/mm.h>
#include <asm/io.h>

  extern maab_dev maab;

  struct vm_operations_struct *ops;

  unsigned long  size;
  struct dma_buffparms *dmap = NULL;

  unsigned long off;
  int i;
	
  maab_says(
  printk("VMA->flags=0x%lx start=0x%x end=0x%lx len=%d offset=0x%lx \n",
	 vma->vm_flags,vma->vm_start,
	 vma->vm_end,
	 vma->vm_end-vma->vm_start,
	 vma->vm_offset);

  printk("maab PCI_ADDR: 0x%lx (0x%lx) \n",
	 maab.pci_base_addr,maab.io_base_addr);	    

  for (i=0;i<0x1C;i+=0x4)   
    printk("Register %02x= %02x %02x %02x %02x\n",
	   i,*(maab.io_base_addr+i+3),*(maab.io_base_addr+i+2),
	   *(maab.io_base_addr+i+1),*(maab.io_base_addr+i));
   /*     printk("Register %02x= %02x %02x %02x %02x\n",
	    i,readb(maab.io_base_addr+i+3),readb(maab.io_base_addr+i+2),
	    readb(maab.io_base_addr+i+1),readb(maab.io_base_addr+i));
	    */
  );
  
  if ( (long) maab.io_base_addr & ~PAGE_MASK )
    {
      printk("maab mmap(): value not on page boundary");
      return -ENXIO;
    }


  if ((off=vma_get_offset (vma)) != 0)
    {
      printk ("maab mmap(): offset must be 0.\n");
      return -EINVAL;
    }

  size = vma_get_end (vma) - vma_get_start (vma);

  if (size > maab.io_size)
    {
      printk ("maab mmap(): size = %ld. Should be %d\n",
	      size, maab.io_size);
      return -EINVAL;
    }

 
  if (remap_page_range (vma_get_start(vma), 
			(unsigned long)phys_to_virt(maab.pci_base_addr),
			vma_get_end (vma) - vma_get_start (vma),
			vma_get_page_prot (vma))){
    printk ("maab mmap(): remap error !\n");
    return -EAGAIN;
  }

  vma_set_inode(vma, inode);
  inode_inc_count(inode);


  maab_says(
  printk("VMA->flags=0x%lx start=0x%x end=0x%lx len=%d offset=0x%lx \n",
	 vma->vm_flags,vma->vm_start,
	 vma->vm_end,
	 vma->vm_end-vma->vm_start,
	 vma->vm_offset);
  );

  }/  /* mmap */

/*
 * This method will make a /proc/module entry
 */

procinfo /{

#include <linux/pci.h>     /* kernel PCI Defs */

  extern maab_dev maab;
  char *p;
	
  p = buf;

  p+=sprintf(p,"MAAB at bus %d index %d with irq-line %d,",
	     maab.pci_bus, maab.pci_index, maab.pci_irq_line);
  p+=sprintf(p,"addr 0x%x\n",maab.pci_base_addr);

  p+=sprintf(p,"   card base address 0x%x\n",
	     maab.pci_base_addr);

  p+=sprintf(p,"   maab memory: %s, %s\n",
	     maab.pci_addr_memtype,maab.pci_addr_prefetch);

  p+=sprintf(p,"   device fn:%d, rev.%d, class_prog 0x%x, class 0x%x\n",
	     maab.pci_device_fn,maab.pci_rev,maab.pci_class_prog,
	     maab.pci_class);

  p+=sprintf(p,"   device cache size %d, latency %d, "
	     "header type %d, bist %d\n",
	     maab.pci_cache_size,maab.pci_latency,
	     maab.pci_header_type,maab.pci_bist);

  p+=sprintf(p,"   status is 0x%x, that means 66MHz %s, UDF %s\n",
	     maab.pci_status,
	     (maab.pci_status&PCI_STATUS_66MHZ)?"yes":"no",
	     (maab.pci_status&PCI_STATUS_UDF)?"1":"0");

  retval = p - buf;		
  }/
}
/* driver maab */ 
